from collections.abc import Callable
from typing import Any

import bson
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema


class PyObjectId(bson.ObjectId):
    @classmethod
    def validate(
        cls, v: bson.ObjectId | bytes, _: core_schema.ValidationInfo = None
    ) -> bson.ObjectId:
        return cls._validate(v)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.with_info_plain_validator_function(
            cls.validate, serialization=core_schema.to_string_ser_schema()
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        schema: core_schema.CoreSchema,
        handler: Callable[[Any], core_schema.CoreSchema],
    ) -> CoreSchema:
        return dict(type="string")

    @classmethod
    def _validate(cls, v: bson.ObjectId | bytes) -> bson.ObjectId:
        if not bson.ObjectId.is_valid(v):
            raise ValueError("Invalid object id")
        return bson.ObjectId(v)



import datetime
from collections.abc import Callable, Generator
from dataclasses import dataclass
from typing import Any, Self

import pendulum
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema

from .helper import get_pydantic_core_schema


class DateTime(pendulum.DateTime):
    """https://stackoverflow.com/a/76719893"""

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], pendulum.DateTime], None, None]:
        yield cls._instance

    @classmethod
    def _instance(cls, v: datetime.datetime | pendulum.DateTime) -> pendulum.DateTime:
        return pendulum.instance(obj=v, tz=pendulum.UTC)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return get_pydantic_core_schema(cls, handler(datetime.datetime))


@dataclass
class DateRange:
    start: DateTime
    end: DateTime

    @classmethod
    def day_of(
        cls,
        dt: pendulum.DateTime,
        tz: pendulum.tz.Timezone | str = pendulum.UTC,
    ) -> Self:
        if dt.tzinfo is None:
            raise ValueError("input datetime must be aware")

        return cls(start=(start := dt.start_of("day").in_tz(tz)), end=start.add(days=1))



import asyncio
import logging
import signal
import threading
import time
import types
from collections.abc import Callable

from .consumer import MessageConsumer
from .signal_handler import SignalHandler, ThreadSignalHandler

logger = logging.getLogger("compose")


class ThreadMessageConsumerRunner:
    def __init__(
        self,
        message_consumer_factory: Callable[[], MessageConsumer],
        signal_handler_factory: Callable[[], SignalHandler] = ThreadSignalHandler,
    ):
        self.message_consumer_factory = message_consumer_factory
        self.signal_handler_factory = signal_handler_factory

        self._received_signal = False
        self._consumers: list[MessageConsumer] = []

        for signum in (signal.SIGINT, signal.SIGTERM):
            signal.signal(signum, self.handle_signal)

    def run(self, num_workers: int = 1) -> None:
        threads = []

        for _ in range(num_workers):
            t = threading.Thread(target=self._run_in_thread)
            t.start()
            threads.append(t)

        while not self._received_signal:
            time.sleep(0.5)

        for consumer in self._consumers:
            consumer.shutdown()

        for t in threads:
            t.join()

    def _run_in_thread(self) -> None:
        with asyncio.Runner() as runner:
            runner.run(self._run_consumer())

    async def _run_consumer(self) -> None:
        message_consumer = self.message_consumer_factory()
        self._consumers.append(message_consumer)
        await message_consumer.run()

    def handle_signal(self, signum: int, _: types.FrameType) -> None:
        logger.info(f"Received {signal.Signals(signum).name}, exiting gracefully")
        self._received_signal = True



import inspect
from collections.abc import Callable
from typing import Annotated, Any

from dependency_injector.wiring import inject
from fastapi import Depends

from compose.dependency.wiring import Provider


def auto_wired[F: Callable[..., Any]](
    provider: Provider,
    *,
    with_injection: bool = False,
) -> Callable[[F], F]:
    """
    FastAPI 엔드포인트에 의존성을 자동으로 주입하는 데코레이터. 해당 데코레이터는 `@inject` 데코레이터보다
    먼저 적용되어야 합니다.

    주의: `__signature__` 속성은 일반 함수에만 존재하므로 해당 데코레이터는 메서드에 사용할 수 없습니다.
    """

    def wrapper(f: F) -> F:
        signature = inspect.signature(f)

        updated_params = []
        for name, param in signature.parameters.items():
            updated_param = param
            try:
                provided = provider(param.annotation)
                updated_param = updated_param.replace(
                    annotation=Annotated[param.annotation, Depends(provided)]
                )
            except ValueError:
                pass

            updated_params.append(updated_param)

        f.__signature__ = signature.replace(parameters=updated_params)

        if with_injection:
            f = inject(f)

        return f

    return wrapper


class AutoWired:
    def __init__(self, provider: Provider, injectors: dict[str, Callable[..., Any]] | None = None):
        self.provider = provider
        self.injectors = injectors or {}

    def __call__[F: Callable[..., Any]](self) -> Callable[[F], F]:
        def wrapper(f: F) -> F:
            signature = inspect.signature(f)

            updated_params = []
            for name, param in signature.parameters.items():
                updated_param = param

                if (injector := self.injectors.get(name)) is not None:
                    updated_param = updated_param.replace(default=Depends(injector))
                else:
                    try:
                        provided = self.provider(param.annotation)
                        updated_param = updated_param.replace(default=Depends(provided))
                    except ValueError:
                        pass

                updated_params.append(updated_param)

            f.__signature__ = signature.replace(parameters=updated_params)
            f = inject(f)

            return f

        return wrapper



import socket
import subprocess
import time
from pathlib import Path
from typing import ClassVar, Self

from testcontainers.mongodb import MongoDbContainer as _MongoDbContainer


class MongoDbContainer(_MongoDbContainer):
    replica_set_name: ClassVar[str] = "rs"
    key_file_path: ClassVar[str] = "/tmp/mongodb-keyfile"
    init_key_file_path: ClassVar[str] = "/docker-entrypoint-initdb.d/01-init-keyfile.sh"

    def __init__(
        self,
        image: str = "mongo:latest",
        port: int = 27017,
        username: str | None = None,
        password: str | None = None,
        dbname: str | None = None,
        **kwargs,
    ):
        super().__init__(
            image=image,
            port=port,
            username=username,
            password=password,
            dbname=dbname,
            **kwargs,
        )
        self._with_replica_set = False

    def get_connection_url(self) -> str:
        connection_url = super().get_connection_url()
        return (
            f"{connection_url}/?replicaSet={self.replica_set_name}"
            if self._with_replica_set
            else connection_url
        )

    def with_replica_set(self) -> Self:
        output = subprocess.run("openssl rand -base64 32".split(), capture_output=True)
        Path(self.key_file_path).write_bytes(output.stdout)
        output = subprocess.run(["chmod", "600", self.key_file_path], capture_output=True)
        if output.returncode:
            raise RuntimeError(output.stderr)

        self.port = self._get_available_port()
        self.with_bind_ports(container=self.port, host=self.port)
        self.with_volume_mapping(host=self.key_file_path, container=self.key_file_path, mode="rw")
        self.with_command(
            f"--keyFile {self.key_file_path} --replSet {self.replica_set_name} --port {self.port}"
        )
        self._with_replica_set = True

        return self

    def _connect(self) -> None:
        super()._connect()
        if self._with_replica_set:
            self._initialize_replica_set()

    def _initialize_replica_set(self) -> None:
        elapsed = 0
        timeout = 30
        interval = 1
        while True:
            exit_code, stdout = self.exec(
                f"mongosh "
                f"--username={self.username} "
                f"--password={self.password} "
                f"--eval=\"rs.initiate({{_id:'{self.replica_set_name}',members:[{{_id:0,host:'localhost:{self.port}'}}]}})\" "
                f"mongodb://localhost:{self.port}"
            )
            if not exit_code:
                break

            time.sleep(interval)
            elapsed += interval
            if elapsed > timeout:
                raise RuntimeError(f"Could not initiate replica set due to {stdout.decode()}")

    @staticmethod
    def _get_available_port() -> int:
        sock = socket.socket()
        sock.bind(("", 0))
        return sock.getsockname()[1]



from .._internal import is_package_installed

if not is_package_installed("pytest"):
    raise ImportError("Install `pytest` to use testing fixtures")

from .enums import *  # noqa: F403
from .fixture import *  # noqa: F403
from .hook import *  # noqa: F403
from .param import *  # noqa: F403



from .._internal import is_package_installed

if not is_package_installed("opentelemetry"):
    raise ImportError("Install `opentelemetry` extra to use opentelemetry features")

from .instrumentation.loguru.instrumentor import LoguruInstrumentor
from .tracer_provider import get_default_tracer_provider

__all__ = ["LoguruInstrumentor", "get_default_tracer_provider"]



from .extra import schema_by_field_name, schema_excludes
from .schema import Error, Id, InvalidParam, ListSchema, Schema, TimeStampedSchema

__all__ = [
    "Error",
    "Id",
    "InvalidParam",
    "ListSchema",
    "Schema",
    "TimeStampedSchema",
    "schema_by_field_name",
    "schema_excludes",
]



from .._internal import is_package_installed

if not is_package_installed("fastapi"):
    raise ImportError("Install `fastapi` to use `compose.fastapi` package")

from .depends import CommandUpdater, UserInjector, create_with_user
from .endpoint import health_check
from .exception_handler import (
    ExceptionHandler,
    ExceptionHandlerInfo,
    create_exception_handler,
)
from .openapi import (
    OpenAPIDoc,
    OpenAPISchema,
    RedocHTML,
    SwaggerUIHTML,
    add_doc_routes,
    additional_responses,
    openapi_tags,
)
from .param import OffsetPaginationParams, WithPath, as_query, to_query, with_depends
from .response import NoContentResponse, ZipStreamingResponse
from .routing import APIRouter
from .security import APIKeyHeader, CookieAuth, HTTPBasic, HTTPBearer, unauthorized_error
from .wiring import AutoWired, auto_wired

__all__ = [
    "APIKeyHeader",
    "APIRouter",
    "AutoWired",
    "CommandUpdater",
    "CookieAuth",
    "ExceptionHandler",
    "ExceptionHandlerInfo",
    "HTTPBasic",
    "HTTPBearer",
    "NoContentResponse",
    "OffsetPaginationParams",
    "OpenAPIDoc",
    "OpenAPISchema",
    "RedocHTML",
    "SwaggerUIHTML",
    "UserInjector",
    "WithPath",
    "ZipStreamingResponse",
    "add_doc_routes",
    "additional_responses",
    "as_query",
    "auto_wired",
    "create_exception_handler",
    "create_with_user",
    "health_check",
    "openapi_tags",
    "to_query",
    "unauthorized_error",
    "with_depends",
]


try:
    from .utils import (  # noqa: F401
        ErrorEvent,
        Level,
        capture_error,
        create_before_send_hook,
        init_sentry,
    )

    __all__.extend(
        [
            "ErrorEvent",
            "Level",
            "capture_error",
            "create_before_send_hook",
            "init_sentry",
        ]
    )
except ImportError:
    pass



from __future__ import annotations

import functools
import warnings
from collections.abc import Iterable
from typing import Any, ClassVar, Self, Unpack, get_args, get_origin, overload

import pendulum
import pymongo
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.results import UpdateResult

from .. import types
from ..entity import Entity
from ..pagination import Pagination
from ..query.mongo import MongoFilterQuery, MongoQuery
from .base import BaseRepository

registry: dict[str, list[pymongo.IndexModel]] = {}


class MongoDocument(dict[str, Any]):
    @classmethod
    def from_entity(cls, entity: Entity, **kwargs: Any) -> Self:
        return entity.model_dump(
            by_alias=True,
            **{key: value for key, value in kwargs.items() if key not in {"by_alias"}},
        )


class MongoRepository[T: Entity](BaseRepository):
    """
    `MongoRepository` 추상 클래스

    상속 인자:
        `session_requirement`: `MongoRepository` 상속 시 `session_requirement` 인자를 통해
        `session` 인자 선언 여부를 검사할 수 있습니다. 프로덕션에서는 `session_requirement`를
        `SessionRequirement.REQUIRED`로 설정하는 것을 권장합니다.

        ```python
        class UserRepository(
            MongoRepository[User],
            session_requirement=SessionRequirement.REQUIRED,
        ):
            ...
        ```

    """

    __collection_name__: ClassVar[str] = ""
    __indexes__: ClassVar[list[pymongo.IndexModel] | None] = None

    def __init__(self, collection: Collection):
        self.collection = collection

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        indexes = cls.__indexes__ or []
        if cls.__collection_name__ not in registry:
            registry[cls.__collection_name__] = indexes
            return

        index_documents = [idx.document for idx in registry[cls.__collection_name__]]
        registry[cls.__collection_name__].extend(
            [idx for idx in indexes if idx.document not in index_documents]
        )

    @classmethod
    def create(cls, database: Database, **kwargs) -> MongoRepository:
        collection = database.get_collection(cls.__collection_name__, **kwargs)
        if (
            cls.__collection_name__ not in database.list_collection_names()
            and cls.__indexes__ is not None
        ):
            collection.create_indexes(cls.__indexes__)

        return cls(collection=collection)

    def find_by_id(
        self,
        entity_id: types.PyObjectId,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | None:
        return self.find_by({"_id": entity_id}, session=session, **kwargs)

    @overload
    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | None: ...

    @overload
    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any],
        session: ClientSession | None = None,
        **kwargs,
    ) -> dict[str, Any] | None: ...

    def find_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> T | dict[str, Any] | None:
        validate_to_entity = projection is None
        query_result = self.collection.find_one(
            filter=filter_,
            session=session,
            **kwargs,
        )
        if query_result is None:
            return None

        return (
            self._entity_type.model_validate(query_result) if validate_to_entity else query_result
        )

    def find_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> dict[str, Any] | None:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return next(query_result, None)

    @overload
    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: None = None,
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[T]: ...

    @overload
    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any],
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[dict[str, Any]]: ...

    def list_by(
        self,
        filter_: dict[str, Any],
        *,
        projection: dict[str, Any] | None = None,
        sort: list[tuple[str, int]] | None = None,
        session: ClientSession | None = None,
        **kwargs,
    ) -> list[T] | list[dict[str, Any]]:
        validate_to_entity = projection is None
        query_result = self.collection.find(
            filter=filter_,
            projection=projection,
            sort=sort,
            session=session,
            **kwargs,
        )
        return (
            [self._entity_type.model_validate(item) for item in query_result]
            if validate_to_entity
            else list(query_result)
        )

    def list_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return list(query_result)

    def add(self, entity: T, session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_one(MongoDocument.from_entity(entity), session=session, **kwargs)

    def add_many(self, entities: list[T], session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_many(
            [MongoDocument.from_entity(entity) for entity in entities], session=session, **kwargs
        )

    def update(self, entity: T, session: ClientSession | None = None, **kwargs) -> None:
        document = MongoDocument.from_entity(entity)
        update_result = self.collection.update_one(
            {"_id": entity.id},
            {"$set": document},
            session=session,
            **kwargs,
        )

        self.on_update(
            update_result=update_result,
            entity=entity,
            session=session,
            **kwargs,
        )

    def on_update(
        self,
        update_result: UpdateResult,
        entity: T,
        session: ClientSession | None = None,
        **kwargs,
    ) -> None:
        if not update_result.modified_count:
            return

        if getattr(entity, "updated_at") is None:
            return

        self.collection.update_one(
            {"_id": entity.id},
            {"$set": {"updated_at": pendulum.DateTime.utcnow()}},
            session=session,
            **kwargs,
        )

    def update_many(
        self, entities: list[T], session: ClientSession | None = None, **kwargs
    ) -> None:
        for entity in entities:
            self.update(entity=entity, session=session, **kwargs)

    def delete(
        self, entity_id: types.PyObjectId, session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.delete_one({"_id": entity_id}, session=session, **kwargs)

    def execute_raw(
        self,
        operation: str,
        session: ClientSession | None = None,
        **operation_kwargs,
    ) -> Any:
        op = getattr(self.collection, operation, None)
        if op is None:
            raise ValueError(f"Unknown operation on collection: {operation}")
        return op(session=session, **operation_kwargs)

    def filter(
        self, qry: MongoFilterQuery, session: ClientSession | None = None, **kwargs
    ) -> Pagination:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        if (unwrapped := next(query_result, None)) is None:
            raise ValueError(f"{qry.__class__.__name__} returned nothing")

        return Pagination(
            total=(unwrapped["metadata"][0]["total"] if unwrapped["metadata"] else 0),
            items=unwrapped["items"],
            page=qry.page,
            per_page=qry.per_page,
        )

    @functools.cached_property
    def _entity_type(self) -> T:
        orig_base = next(
            (base for base in self.__class__.__orig_bases__ if get_origin(base) is MongoRepository),
            None,
        )
        if orig_base is None:
            raise ValueError("No origin base found")

        return get_args(orig_base)[0]


def setup_indexes(
    db_names: Iterable[str],
    mongo_client: pymongo.MongoClient | None = None,
    mongo_uri: str | None = None,
) -> None:
    # NOTE: 동일한 이름의 컬렉션이 여러 DB에 존재할 경우, 컬렉션별로 인덱스를 생성할 수 없음

    if mongo_client is None and mongo_uri is None:
        raise ValueError("Either `mongo_client` or `mongo_uri` must be provided")

    if mongo_uri is not None:
        warnings.warn(
            (
                "`mongo_uri` is deprecated and will be removed in the future. "
                "Use `mongo_client` instead."
            ),
            DeprecationWarning,
        )

        with pymongo.MongoClient(mongo_uri) as mongo_client:
            setup_database_indexes(*(mongo_client.get_database(db_name) for db_name in db_names))

        return

    setup_database_indexes(*(mongo_client.get_database(db_name) for db_name in db_names))


def setup_database_indexes(*databases: Unpack[tuple[Database, ...]]) -> None:
    for database in databases:
        collection_names = database.list_collection_names()
        for collection_name, indexes in registry.items():
            if not collection_name or collection_name not in collection_names:
                continue
            collection = database.get_collection(collection_name)

            previous_index_names = {index["name"] for index in collection.list_indexes()}
            current_index_names = {
                *(index.document["name"] for index in indexes),
                "_id_",
            }

            for index_name in previous_index_names - current_index_names:
                collection.drop_index(index_name)

            if indexes:
                collection.create_indexes(indexes)



from .. import container, field, types


class Event(container.BaseModel):
    id: types.PyObjectId = field.IdField(default_factory=types.PyObjectId)
    published_at: types.DateTime = field.DateTimeField()



from .._internal import is_package_installed

if not is_package_installed("httpx"):
    raise ImportError("Install `httpx` extra to use httpx features")

from .auth.header import HeaderAuth

__all__ = ["HeaderAuth"]



import abc
import functools

from ..uow import MongoUnitOfWork


class MongoTransactionHandler[T, R](abc.ABC):
    def __init__(self, uow: MongoUnitOfWork):
        self.uow = uow

    def handle(self, message: T, /) -> R:
        return self.uow.with_transaction(functools.partial(self.handle_in_transaction, message))

    def handle_in_transaction(self, message: T) -> R:
        raise NotImplementedError



from fastapi import APIRouter as _FastAPIRouter

_OVERRIDE_DEFAULTS = {"response_model_by_alias": False}
_OVERRIDE_METHODS = {
    "get",
    "post",
    "put",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
    "api_route",
    "add_api_route",
}

APIRouter = _FastAPIRouter

for method_name in _OVERRIDE_METHODS:
    func = getattr(APIRouter, method_name)
    func.__kwdefaults__ = func.__kwdefaults__ | _OVERRIDE_DEFAULTS



from __future__ import annotations

from typing import Any, Self, TypeVar

from pydantic.alias_generators import to_camel

DictExpression = dict[str, Any]
ListExpression = list[DictExpression]
Expression = TypeVar("Expression", DictExpression, ListExpression)

Id = "$_id"


class MongoKeyword(str):
    def __new__(cls, v: str):
        if v != _camelize(v):
            raise ValueError(f"Cannot interpret {v} as valid mongo keyword")

        return super().__new__(cls, v)

    @classmethod
    def from_py(cls, v: str) -> Self:
        return cls(_camelize(v))


def _camelize(v: str) -> str:
    return to_camel(v.strip("_"))


class _FieldPath(str):
    def __new__(cls, v: str):
        if not v.startswith("$"):
            raise ValueError("path must be prefixed with $")

        return super().__new__(cls, v)


class _String(str):
    def __new__(cls, v: str):
        if v.startswith("$"):
            raise ValueError("string must not be prefixed with $")

        return super().__new__(cls, v)



from typing import Any

from ..base import Evaluable, GeneralAggregationOperator, Operator
from ..types import DictExpression


class IfNull(GeneralAggregationOperator):
    mongo_operator = "$ifNull"


class Cond(Operator):
    def __init__(self, if_: Any, then: Any, else_: Any):
        """
        Aggregation operator `$cond`

        Reference:
            https://www.mongodb.com/docs/manual/reference/operator/aggregation/cond/#mongodb-expression-exp.-cond
        """

        self.if_ = if_
        self.then = then
        self.else_ = else_

    def expression(self) -> DictExpression:
        return {
            "$cond": {
                "if": Evaluable(self.if_).expression(),
                "then": Evaluable(self.then).expression(),
                "else": Evaluable(self.else_).expression(),
            }
        }



from .base import OffsetPaginationQuery, Query
from .mongo.query import MongoFilterQuery, MongoOffsetFilterQuery, MongoQuery

__all__ = [
    "MongoFilterQuery",
    "MongoOffsetFilterQuery",
    "MongoQuery",
    "OffsetPaginationQuery",
    "Query",
]



from __future__ import annotations

import urllib.parse
from typing import Any, ClassVar, Self, cast

from .primitive import Str


class S3ContentUrl(Str):
    base_url: ClassVar[str]

    def __new__(cls, v: Any):
        if getattr(cls, "base_url", None) is None:
            raise ValueError("`base_url` must be set")

        v = urllib.parse.unquote(str(v).strip("/"))
        if v.startswith(cls.base_url):
            v = v[len(cls.base_url) :].lstrip("/")

        quoted_parts = [urllib.parse.quote(part, safe="~()*!.'") for part in v.split("/")]
        url = f"{cls.base_url}/{'/'.join(quoted_parts)}"
        return super().__new__(cls, url)

    @classmethod
    def with_base_url(cls, base_url: str) -> type[S3ContentUrl]:
        return cast(
            type[Self],
            type(
                cls.__name__,
                (cls,),
                {"base_url": base_url},
            ),
        )



from pydantic import ConfigDict

from .. import container, schema, types


class Command(container.BaseModel): ...


class UserCommand(Command):
    user_id: types.PyObjectId | None = None

    model_config = ConfigDict(json_schema_extra=schema.extra.schema_excludes("user_id"))



from collections.abc import Iterable
from typing import Any, Self

from .base import Evaluable, Merge, Operator, Stage
from .evaulation import Expr
from .logical import And, Nor, Or
from .pipeline import Pipeline
from .sort import SortBy
from .types import DictExpression, ListExpression, MongoKeyword, _FieldPath


class EmptyStage(Stage[DictExpression]):
    def expression(self) -> DictExpression:
        return {}


class Match(Stage[DictExpression]):
    def __init__(self, op: Operator):
        self.op = op

    def expression(self) -> DictExpression:
        if not (expression := self.op.expression()):
            return EmptyStage().expression()
        return {"$match": expression}

    @classmethod
    def and_(cls, *ops: Operator) -> Self:
        return cls(And(*ops))

    @classmethod
    def or_(cls, *ops: Operator) -> Self:
        return cls(Or(*ops))

    @classmethod
    def nor(cls, *ops: Operator) -> Self:
        return cls(Nor(*ops))

    @classmethod
    def expr(cls, op: Any) -> Self:
        return cls(Expr(op))


class Sort(Stage[DictExpression]):
    def __init__(self, *criteria: SortBy):
        self.criteria = list(criteria)

    def expression(self) -> DictExpression:
        if not (merged := Merge.dict(*self.criteria).expression()):
            raise ValueError("Expression cannot be empty")
        return {"$sort": merged}


class Spec(Operator):
    def __init__(self, field: str, spec: Any):
        self.field = field
        self.spec = spec

    def expression(self) -> DictExpression:
        return {self.field: Evaluable(self.spec).expression()}

    @classmethod
    def include(cls, field: str) -> Self:
        return cls(field=field, spec=1)

    @classmethod
    def exclude(cls, field: str) -> Self:
        return cls(field=field, spec=0)

    @classmethod
    def ref(cls, field: str, spec: str) -> Self:
        return cls(field=field, spec=_FieldPath(spec))


class Project(Stage[DictExpression]):
    def __init__(self, *specs: Spec):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$project": Merge.dict(*self.specs).expression()}

    @classmethod
    def from_attrs(cls, **attrs: Any) -> Self:
        return cls(*(Spec(field, spec) for field, spec in attrs.items()))

    @classmethod
    def from_(
        cls,
        include: Iterable[str] | None = None,
        exclude: Iterable[str] | None = None,
        ref: dict[str, Any] | None = None,
    ) -> Self:
        if include is None and exclude is None and ref is None:
            raise ValueError("At least one of `include`, `exclude`, or `ref` must be provided")

        return cls(
            *(Spec.include(field) for field in (include or {})),
            *(Spec.exclude(field) for field in (exclude or {})),
            *(Spec(field, spec) for field, spec in (ref or {}).items()),
        )


class Lookup(Stage[DictExpression]):
    def __init__(self, from_: str, as_: str | None = None):
        self.from_ = from_
        self.as_ = as_ or self.from_

    def expression(self) -> DictExpression:
        return {
            "$lookup": {
                MongoKeyword.from_py(field): Evaluable(value).expression()
                for field, value in self.__dict__.items()
            }
        }


class MatchLookup(Lookup):
    def __init__(self, from_: str, local_field: str, foreign_field: str, as_: str | None = None):
        super().__init__(from_=from_, as_=as_)
        self.local_field = local_field
        self.foreign_field = foreign_field

    @classmethod
    def on_id(cls, from_: str, local_field: str, as_: str | None = None) -> Self:
        return cls(
            from_=from_,
            local_field=local_field,
            foreign_field="_id",
            as_=as_,
        )


class SubqueryLookup(Lookup):
    def __init__(
        self,
        from_: str,
        pipeline: Pipeline,
        let: Operator | None = None,
        as_: str | None = None,
    ):
        super().__init__(from_=from_, as_=as_)
        self.let = let
        self.pipeline = pipeline


class Unwind(Stage[DictExpression]):
    def __init__(
        self,
        path: str,
        include_array_index: str | None = None,
        preserve_null_and_empty_arrays: bool | None = None,
    ):
        self.path = _FieldPath(path)
        self.include_array_index = include_array_index
        self.preserve_null_and_empty_arrays = preserve_null_and_empty_arrays

    def expression(self) -> DictExpression:
        return {
            "$unwind": {
                MongoKeyword.from_py(field): value
                for field, value in self.__dict__.items()
                if value is not None
            }
        }

    @classmethod
    def preserve_missing(cls, path: str, include_array_index: str | None = None) -> Self:
        return cls(
            path=path,
            include_array_index=include_array_index,
            preserve_null_and_empty_arrays=True,
        )


class Set(Stage[DictExpression]):
    def __init__(self, *specs: Spec):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$set": Merge.dict(*self.specs).expression()}


class FacetSubPipeline(Operator):
    def __init__(self, output_field: str, pipeline: Pipeline):
        self.output_field = output_field
        self.pipeline = pipeline

    def expression(self) -> DictExpression:
        return {self.output_field: self.pipeline.expression()}


class Facet(Stage[DictExpression]):
    def __init__(self, *pipelines: FacetSubPipeline):
        self.pipelines = list(pipelines)

    def expression(self) -> DictExpression:
        return {"$facet": Merge.dict(*self.pipelines).expression()}


class Skip(Stage[DictExpression]):
    def __init__(self, skip: int):
        self.skip = skip

    def expression(self) -> DictExpression:
        return {"$skip": self.skip}


class Limit(Stage[DictExpression]):
    def __init__(self, limit: int):
        self.limit = limit

    def expression(self) -> DictExpression:
        return {"$limit": self.limit}


class TextSearchOperator(Operator):
    def __init__(self, query: str, path: str | list[str]):
        self.query = query
        self.path = path

    def expression(self) -> DictExpression:
        return {"text": {"query": self.query, "path": self.path}}


class Search(Stage[DictExpression]):
    def __init__(self, index: str, op: Operator):
        self.index = index
        self.op = op

    def expression(self) -> DictExpression:
        return {"$search": {"index": self.index} | self.op.expression()}


class Pagination(Stage[DictExpression]):
    def __init__(
        self,
        page: int | None = None,
        per_page: int | None = None,
        metadata_ops: list[Operator] | None = None,
    ):
        if not ((page is not None and per_page is not None) or (page is None and per_page is None)):
            raise ValueError("`page` and `per_page` are mutual inclusive")

        self.page = page
        self.per_page = per_page
        self.metadata_ops = metadata_ops or []

        self.can_paginate = self.page is not None and self.per_page is not None

    def expression(self) -> DictExpression:
        return {
            "$facet": {
                "metadata": [{"$count": "total"}, *[op.expression() for op in self.metadata_ops]],
                "items": self._pagination_expression(),
            }
        }

    def _pagination_expression(self) -> ListExpression:
        if not self.can_paginate:
            return []

        return [
            Skip((self.page - 1) * self.per_page).expression(),  # type: ignore
            Limit(self.per_page).expression(),  # type: ignore
        ]


class ReplaceRoot(Stage[DictExpression]):
    def __init__(self, new_root: Any):
        self.new_root = new_root

    def expression(self) -> DictExpression:
        return {"$replaceRoot": {"newRoot": Evaluable(self.new_root).expression()}}


class Group(Stage[DictExpression]):
    def __init__(self, *ops: Operator, key: Any | None = None):
        self.ops = list(ops)
        self.key = key

    def expression(self) -> DictExpression:
        return {
            "$group": {
                "_id": Evaluable(self.key).expression(),
                **Merge.dict(*self.ops).expression(),
            }
        }

    @classmethod
    def by_null(cls, *ops: Operator) -> Self:
        return cls(*ops, key=None)


class Sample(Stage[DictExpression]):
    def __init__(self, size: int):
        self.size = size

    def expression(self) -> DictExpression:
        return Evaluable({"$sample": {"size": self.size}}).expression()



from collections.abc import Iterable
from typing import Self

from dependency_injector import containers

from .wiring import Wirer


class DeclarativeContainer(containers.DeclarativeContainer):
    @classmethod
    def wired(
        cls,
        wirer: Wirer,
        modules: Iterable[str] | None = None,
        from_package: str | None = None,
    ) -> Self:
        container = cls()
        wirer(container, modules=modules, from_package=from_package)
        return container



from __future__ import annotations

import copy
import inspect
from collections.abc import Callable, Generator
from typing import Any, Protocol, get_args

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class SupportsGetValidators(Protocol):
    """Naming Convention: https://github.com/python/typeshed/issues/4174"""

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], Any], None, None]: ...


def chain(*validators: Callable[[Any], Any]) -> Callable[[Any], Any]:
    def apply_chain(v: Any) -> Any:
        result = copy.deepcopy(v)
        for _validator in validators:
            result = _validator(result)
        return result

    return apply_chain


def get_pydantic_core_schema(
    t: type[SupportsGetValidators], schema: core_schema.CoreSchema, /
) -> core_schema.CoreSchema:
    """Pydantic v1 커스텀 타입에 구현한 검증 제너레이터를 v2 커스텀 타입 검증 함수로 변환합니다.

    class CustomType(str):
        @classmethod
        def __get_validators__(cls):
            yield str_validator

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> core_schema.CoreSchema:
            return compose.types.get_pydantic_core_schema(cls, handler(str))
    """

    if not hasattr(t, "__get_validators__"):
        type_name = t.__name__ if inspect.isclass(t) else t.__class__.__name__
        raise AttributeError(f"`{type_name}` does not have `__get_validators__` method")

    return core_schema.no_info_after_validator_function(chain(*t.__get_validators__()), schema)


class CoreSchemaGettable[T]:
    """Pydantic v1 커스텀 타입을 v2 커스텁 타입으로 변환하는 믹스인

    ```python
    class CustomType(str, compose.types.CoreSchemaGettable[str]):
        @classmethod
        def __get_validators__(cls):
            yield str_validator
    ```

    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        core_schema_gettable_cls = next(
            (base for base in source_type.__orig_bases__ if base.__name__ == "CoreSchemaGettable"),
            None,
        )
        if core_schema_gettable_cls is None:
            raise AttributeError(f"`{source_type.__name__}` does not inherit `CoreSchemaGettable`")

        validatable_type = get_args(core_schema_gettable_cls)[0]
        return get_pydantic_core_schema(cls, handler(validatable_type))



import asyncio
from collections.abc import Awaitable, Callable, Hashable
from typing import Literal, overload

from pydantic.dataclasses import dataclass


class AsyncJob[K: Hashable, **P, T]:
    def __init__(
        self,
        key: K,
        func: Callable[P, Awaitable[T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        self.key = key
        self.func = func
        self.args = args
        self.kwargs = kwargs


@dataclass
class Result[K: Hashable, T]:
    key: K
    value: T


class AsyncTaskExecutor:
    def __init__(self, concurrency: int, timeout: float | None = None):
        self.concurrency = concurrency
        self.timeout = timeout

    @overload
    async def execute[K, **P, T](
        self,
        jobs: list[AsyncJob[K, P, T]],
        group: Literal[True] = True,
        concurrency: int | None = None,
        timeout: float | None = None,
    ) -> dict[K, Result[K, T]]: ...

    @overload
    async def execute[K, **P, T](
        self,
        jobs: list[AsyncJob[K, P, T]],
        group: Literal[False] = False,
        concurrency: int | None = None,
        timeout: float | None = None,
    ) -> list[Result[K, T]]: ...

    async def execute[K, **P, T](
        self,
        jobs: list[AsyncJob[K, P, T]],
        group: bool = False,
        concurrency: int | None = None,
        timeout: float | None = None,
    ) -> list[Result[K, T]] | dict[K, Result[K, T]]:
        result = await self._execute(
            jobs=jobs,
            concurrency=concurrency,
            timeout=timeout,
        )

        return {item.key: item for item in result} if group else result

    async def _execute[K, **P, T](
        self,
        jobs: list[AsyncJob[K, P, T]],
        concurrency: int | None = None,
        timeout: float | None = None,
    ) -> list[Result[K, T]]:
        concurrency = concurrency if concurrency is not None else self.concurrency
        timeout = timeout if timeout is not None else self.timeout

        coro = execute_jobs(jobs=jobs, semaphore=asyncio.Semaphore(concurrency))

        if timeout is None:
            return await coro

        async with asyncio.timeout(timeout):
            return await coro


async def execute_jobs[K, **P, T](
    jobs: list[AsyncJob[K, P, T]],
    semaphore: asyncio.Semaphore,
) -> list[Result[K, T]]:
    try:
        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(execute_job(job=job, semaphore=semaphore)) for job in jobs]
    except* Exception as eg:
        raise eg.exceptions[0]

    return [task.result() for task in tasks]


async def execute_job[K, **P, T](
    job: AsyncJob[K, P, T], semaphore: asyncio.Semaphore
) -> Result[K, T]:
    async with semaphore:
        result = await job.func(*job.args, **job.kwargs)

    return Result(key=job.key, value=result)



import math
from typing import Any

from pydantic import Field

from . import container


class Pagination(container.BaseModel):
    items: list[Any]
    total: int
    page: int | None = Field(None, ge=1)
    per_page: int | None = Field(None, ge=1)
    extra: dict[str, Any] = Field(default_factory=dict)

    @property
    def prev_page(self) -> int | None:
        return self.page - 1 if self.has_prev else None  # type: ignore

    @property
    def next_page(self) -> int | None:
        return self.page + 1 if self.has_next else None  # type: ignore

    @property
    def has_prev(self) -> bool:
        return self.page is not None and self.page > 1

    @property
    def has_next(self) -> bool:
        return self.page is not None and self.page < self.pages

    @property
    def pages(self) -> int:
        return math.ceil(self.total / self.per_page) if self.per_page is not None else 1

    @property
    def is_empty(self) -> bool:
        return not self.total



from collections.abc import Callable, Iterable
from typing import TypeAlias

from .base import Merge, Operator
from .pipeline import Pipeline as _Pipeline
from .raw import Raw
from .types import DictExpression, ListExpression

Expressionable: TypeAlias = Operator | DictExpression | ListExpression


class _Map[T]:
    def __init__(self, collection: Iterable[T], callback: Callable[[T], Operator]):
        self.collection = collection
        self.callback = callback

    def eval(self) -> list[Operator]:
        return [self.callback(item) for item in self.collection]


def Map[T](collection: Iterable[T], callback: Callable[[T], Operator]) -> list[Operator]:
    return _Map(collection, callback).eval()


class _Filter:
    def __init__(self, collection: list[Operator], predicate: Callable[[Operator], bool]):
        self.collection = collection
        self.predicate = predicate

    def eval(self) -> list[Operator]:
        return [item for item in self.collection if self.predicate(item)]


def Filter(collection: list[Operator], predicate: Callable[[Operator], bool]) -> list[Operator]:
    return _Filter(collection, predicate=predicate).eval()


class NonEmpty:
    def __call__(self, op: Operator) -> bool:
        return bool(op.expression())


class _Flatten:
    def __init__(self, ops: Iterable[Expressionable]):
        self.ops = list(ops)

    def eval(self) -> list[Operator]:
        result = []
        for op in self.ops:
            exp = op.expression() if isinstance(op, Operator) else op

            match exp:
                case dict():
                    result.append(Raw(exp))
                case list():
                    for e in exp:
                        result.extend([Raw(e)] if isinstance(e, dict) else self.__class__(e).eval())
                case _:
                    raise ValueError(f"Expression must be dict or list, not {type(exp)} ({exp})")

        return result


def Flatten(ops: Iterable[Expressionable]) -> list[Operator]:
    return _Flatten(ops).eval()


def Q(*ops: *tuple[Operator, ...]) -> DictExpression:
    return Merge.dict(*ops).expression()


def Pipeline(*ops: Operator) -> ListExpression:
    return _Pipeline(*ops).expression()



from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Attributes, LabelValue, Resource
from opentelemetry.sdk.trace import (
    ConcurrentMultiSpanProcessor,
    IdGenerator,
    SpanLimits,
    SynchronousMultiSpanProcessor,
    TracerProvider,
)
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk.trace.sampling import Sampler


class ServiceResourceAttrs:
    def __init__(
        self,
        name: str,
        version: str,
        environment: str,
        **extra: LabelValue,
    ):
        self._attrs = {
            "name": name,
            "version": version,
            "environment": environment,
            **extra,
        }

    def to_attrs(self) -> Attributes:
        return {f"service.{key}": value for key, value in self._attrs.items()}


def get_default_tracer_provider(
    resource_attrs: ServiceResourceAttrs,
    exporter_endpoint: str,
    sampler: Sampler | None = None,
    shutdown_on_exit: bool = True,
    active_span_processor: (
        SynchronousMultiSpanProcessor | ConcurrentMultiSpanProcessor | None
    ) = None,
    id_generator: IdGenerator | None = None,
    span_limits: SpanLimits | None = None,
) -> TracerProvider:
    tracer_provider = TracerProvider(
        resource=Resource.create(resource_attrs.to_attrs()),
        sampler=sampler,
        shutdown_on_exit=shutdown_on_exit,
        active_span_processor=active_span_processor,
        id_generator=id_generator,
        span_limits=span_limits,
    )
    exporter = OTLPSpanExporter(endpoint=exporter_endpoint)
    tracer_provider.add_span_processor(BatchSpanProcessor(exporter))
    return tracer_provider



from typing import Any, ClassVar

from . import container, field, types


class Entity(container.TimeStampedModel):
    id: types.PyObjectId = field.IdField(default_factory=types.PyObjectId)

    updatable_fields: ClassVar[set[str]] = set()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        super().__pydantic_init_subclass__(**kwargs)

        fields = set(cls.model_fields.keys())
        if diff := set(cls.updatable_fields) - fields:
            raise ValueError(f"`updatable_fields` must be subset of {fields}, but got {diff}")

    def update(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if key not in self.updatable_fields:
                continue

            setattr(self, key, value)



from .byte_unit import Byte
from .datetime import DateRange, DateTime
from .duration import MilliSeconds, Seconds
from .helper import CoreSchemaGettable, SupportsGetValidators, chain, get_pydantic_core_schema
from .object_id import PyObjectId
from .primitive import Float, Int, IntList, List, Str, StrList, TypedList, validator
from .url import S3ContentUrl
from .web import ContentDisposition

__all__ = [
    "Byte",
    "ContentDisposition",
    "CoreSchemaGettable",
    "DateRange",
    "DateTime",
    "Float",
    "Int",
    "IntList",
    "List",
    "MilliSeconds",
    "PyObjectId",
    "S3ContentUrl",
    "Seconds",
    "Str",
    "StrList",
    "SupportsGetValidators",
    "TypedList",
    "chain",
    "get_pydantic_core_schema",
    "validator",
]



import abc
from typing import Any

from ..base import OffsetPaginationQuery, Query


class MongoQuery(Query, abc.ABC):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


class MongoFilterQuery(OffsetPaginationQuery, MongoQuery):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


MongoOffsetFilterQuery = MongoFilterQuery



import enum
from collections.abc import Callable
from typing import Any

from fastapi import Depends, FastAPI, Response
from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
from fastapi.responses import HTMLResponse

from compose import container, schema


def openapi_tags(tag: type[enum.StrEnum]) -> list[dict[str, Any]]:
    return [{"name": member.value} for member in tag.__members__.values()]


def additional_responses(
    *status_codes: int, schema_type: type[container.BaseModel] = schema.Error
) -> dict[int, dict[str, Any]]:
    return {int(status_code): {"model": schema_type} for status_code in sorted(status_codes)}


class OpenAPIDoc:
    def add_to_app(self, app: FastAPI) -> None:
        raise NotImplementedError

    def get_endpoint(self, app: FastAPI) -> Callable[[], Response]:
        raise NotImplementedError


class SwaggerUIHTML(OpenAPIDoc):
    def __init__(
        self,
        path: str = "/docs",
        dependencies: list[Depends] | None = None,
        openapi_url: str = "/openapi.json",
        **kwargs: Any,
    ):
        self.path = path
        self.dependencies = dependencies
        self.openapi_url = openapi_url
        self.kwargs = kwargs

    def add_to_app(self, app: FastAPI) -> None:
        app.get(
            self.path,
            response_class=HTMLResponse,
            dependencies=self.dependencies,
            include_in_schema=False,
        )(self.get_endpoint(app))

    def get_endpoint(self, app: FastAPI) -> Callable[[], HTMLResponse]:
        def endpoint():
            default_kwargs = {
                "openapi_url": app.root_path + self.openapi_url,
                "title": f"{app.title} - Swagger UI",
                "oauth2_redirect_url": (
                    app.swagger_ui_oauth2_redirect_url
                    and app.root_path + app.swagger_ui_oauth2_redirect_url
                ),
                "init_oauth": app.swagger_ui_init_oauth,
                "swagger_ui_parameters": app.swagger_ui_parameters,
            }
            return get_swagger_ui_html(**(default_kwargs | self.kwargs))

        return endpoint


class RedocHTML(OpenAPIDoc):
    def __init__(
        self,
        path: str = "/redoc",
        dependencies: list[Depends] | None = None,
        openapi_url: str = "/openapi.json",
        **kwargs: Any,
    ):
        self.path = path
        self.dependencies = dependencies
        self.openapi_url = openapi_url
        self.kwargs = kwargs

    def add_to_app(self, app: FastAPI) -> None:
        app.get(
            self.path,
            response_class=HTMLResponse,
            dependencies=self.dependencies,
            include_in_schema=False,
        )(self.get_endpoint(app))

    def get_endpoint(self, app: FastAPI) -> Callable[[], Response]:
        def endpoint():
            default_kwargs = {
                "openapi_url": app.root_path + self.openapi_url,
                "title": f"{app.title} - ReDoc",
            }
            return get_redoc_html(**(default_kwargs | self.kwargs))

        return endpoint


class OpenAPISchema(OpenAPIDoc):
    def __init__(self, path: str = "/openapi.json"):
        self.path = path

    def add_to_app(self, app: FastAPI) -> None:
        app.get(
            self.path,
            include_in_schema=False,
        )(self.get_endpoint(app))

    def get_endpoint(self, app: FastAPI, **kwargs: Any) -> Callable[[], dict[str, Any]]:
        def endpoint() -> dict[str, Any]:
            return app.openapi()

        return endpoint


def add_doc_routes(app: FastAPI, docs: list[OpenAPIDoc], cond: bool) -> None:
    if not cond:
        return

    for doc in docs:
        doc.add_to_app(app)



from typing import Any, Self

from .. import constants
from . import primitive


class Byte(primitive.Int):
    def __new__(cls, v: Any, /) -> Self:
        v = super().__new__(cls, v)
        if v < 0:
            raise ValueError(f"`{cls.__name__}` must be a non-negative integer")
        return v

    @classmethod
    def from_mib(cls, v: int | float, /) -> Self:
        return cls(v * constants.Unit.MIB)

    @classmethod
    def from_gib(cls, v: int | float, /) -> Self:
        return cls(v * constants.Unit.GIB)



import collections
import concurrent.futures
import functools
from collections.abc import Callable, Hashable
from typing import Literal, Self, overload

from pydantic.dataclasses import dataclass


def execute_in_pool[K, T](
    pool_factory: Callable[[], concurrent.futures.Executor],
    funcs: dict[K, functools.partial[T]],
    timeout: int | None = None,
) -> dict[K, T]:
    result = {}
    with pool_factory() as executor:
        future_to_key = dict()
        for key, func in funcs.items():
            future = executor.submit(func)
            future_to_key[future] = key

        for future in concurrent.futures.as_completed(future_to_key, timeout=timeout):
            result[future_to_key[future]] = future.result()

    return result


class ThreadPoolJob[K: Hashable, **P, T]:
    def __init__(
        self,
        key: K,
        func: Callable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        self.key = key
        self.func = func
        self.args = args
        self.kwargs = kwargs


@dataclass
class Result[K: Hashable, T]:
    key: K
    value: T


class ThreadPoolExecutor:
    def __init__(self, max_workers: int, timeout: int | None = None):
        self.max_workers = max_workers
        self.timeout = timeout

    @overload
    def execute[K, **P, T](
        self,
        jobs: list[ThreadPoolJob[K, P, T]],
        group: Literal[True] = True,
        max_workers: int | None = None,
        timeout: int | None = None,
    ) -> dict[K, Result[K, T]]: ...

    @overload
    def execute[K, **P, T](
        self,
        jobs: list[ThreadPoolJob[K, P, T]],
        group: Literal[False] = False,
        max_workers: int | None = None,
        timeout: int | None = None,
    ) -> list[Result[K, T]]: ...

    def execute[K, **P, T](
        self,
        jobs: list[ThreadPoolJob[K, P, T]],
        group: bool = False,
        max_workers: int | None = None,
        timeout: int | None = None,
    ) -> list[Result[K, T]] | dict[K, Result[K, T]]:
        result = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_workers or self.max_workers
        ) as executor:
            future_to_key = {}

            for job in jobs:
                future = executor.submit(job.func, *job.args, **job.kwargs)
                future_to_key[future] = job.key

            for future in concurrent.futures.as_completed(
                future_to_key, timeout=timeout or self.timeout
            ):
                result.append(Result(key=future_to_key[future], value=future.result()))

        return {result.key: result for result in result} if group else result


class DependencyJob[K: Hashable, **P, T]:
    def __init__(
        self,
        key: K,
        dependencies: set[K],
        func: Callable[P, T],
        **kwargs: P.kwargs,
    ):
        self.key = key
        self.dependencies = dependencies
        self.func = func
        self.kwargs = kwargs

    @classmethod
    def no_dependencies(
        cls,
        key: K,
        func: Callable[P, T],
        **kwargs: P.kwargs,
    ) -> Self:
        return cls(
            key=key,
            dependencies=set(),
            func=func,
            **kwargs,
        )


class DependencyExecutor:
    def __init__(self, max_workers: int, timeout: int | None = None):
        self.max_workers = max_workers
        self.timeout = timeout

    def execute[K, **P, T](
        self,
        jobs: list[DependencyJob[K, P, T]],
        max_workers: int | None = None,
        timeout: int | None = None,
    ) -> dict[K, T]:
        if not jobs:
            return {}

        max_workers = max_workers or self.max_workers

        job_map = {job.key: job for job in jobs}
        dependencies = {job.key: job.dependencies for job in jobs}
        dependents = self._build_dependents(dependencies)

        completed = {}
        ready_queue = self._get_ready_queue(jobs)

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            active_futures = {}

            while len(completed) < len(jobs):
                while ready_queue and len(active_futures) < max_workers:
                    job_key = ready_queue.popleft()
                    job = job_map[job_key]
                    future = executor.submit(job.func, **job.kwargs)
                    active_futures[future] = job_key

                if not active_futures:
                    continue

                done, _ = concurrent.futures.wait(
                    active_futures,
                    return_when=concurrent.futures.FIRST_COMPLETED,
                    timeout=timeout or self.timeout,
                )

                for future in done:
                    job_id = active_futures.pop(future)
                    completed[job_id] = future.result()

                    # 완료된 작업에 의존하는 작업 중 모든 의존성이 충족된 작업만 확인
                    # dependencies = {"A": set(), "B": set(), "C": {"A"}, "D": {"A", "B"}, "E": {"C", "D"}}
                    # dependents = {"A": {"C", "D"}, "B": {"D"}, "C": {"E"}, "D": {"E"}}
                    # 1. completed = {}, ready_queue = [A, B]
                    # 2. A 완료 -> completed = {"A": "result_A"}, dependents.get("A") = {"C", "D"}
                    #    dependencies["C"] = {"A"} ⊆ completed.keys() -> ready_queue.append("C")
                    #    dependencies["D"] = {"A", "B"} ⊈ completed.keys() -> X
                    # 3. B 완료 -> completed = {"A": "result_A", "B": "result_B"}, dependents.get("B") = {"D"}
                    #    dependencies["D"] = {"A", "B"} ⊆ completed.keys() -> ready_queue.append("D")
                    # C, D, E도 같은 방식으로 처리
                    for dependent_id in dependents.get(job_id, set()):
                        if dependent_id not in completed and dependencies[dependent_id].issubset(
                            completed.keys()
                        ):
                            ready_queue.append(dependent_id)

        return completed

    @staticmethod
    def _get_ready_queue[K, **P, T](jobs: list[DependencyJob[K, P, T]]) -> collections.deque[K]:
        ready_queue = collections.deque()
        for job in jobs:
            if not job.dependencies:
                ready_queue.append(job.key)
        return ready_queue

    @staticmethod
    def _build_dependents[K](dependencies: dict[str, set[K]]) -> dict[str, set[K]]:
        dependents = collections.defaultdict(set)
        for key, deps in dependencies.items():
            for dep in deps:
                dependents[dep].add(key)
        return dict(dependents)



from .._internal import is_package_installed

if not is_package_installed("testcontainers"):
    raise ImportError("Install `testcontainers` to use testing fixtures")

from .mongodb import MongoDbContainer

__all__ = ["MongoDbContainer"]



from __future__ import annotations

import json
from typing import Any, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict
from typing_extensions import Self

from . import field, types

type AbstractSetIntStr = set[int] | set[str]
type MappingIntStrAny = dict[int, Any] | dict[str, Any]

type IncEx = set[int] | set[str] | dict[int, Any] | dict[str, Any] | None
Model = TypeVar("Model", bound=PydanticBaseModel)


class BaseModel(PydanticBaseModel):
    @classmethod
    def from_model(cls, model: BaseModel) -> Self:
        return cls.model_validate(model.model_dump())

    def json(
        self,
        indent: int | None = None,
        include: IncEx = None,
        exclude: IncEx = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> str:
        return self.model_dump_json(
            indent=indent,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=True,
        )

    def dict(
        self,
        include: IncEx = None,
        exclude: IncEx = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ):
        return self.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=True,
        )

    def copy(
        self,
        *,
        include: AbstractSetIntStr | MappingIntStrAny | None = None,
        exclude: AbstractSetIntStr | MappingIntStrAny | None = None,
        update: dict[str, Any] | None = None,
        deep: bool = False,
        validate: bool = False,
    ) -> Model:
        result = super().model_copy(update=update, deep=deep)
        if not validate:
            return result

        return self.__class__.model_validate(result.model_dump_json(warnings=False))

    def model_copy(
        self,
        *,
        update: dict[str, Any] | None = None,
        deep: bool = False,
        validate: bool = False,
    ) -> Self:
        result = super().model_copy(update=update, deep=deep)
        if not validate:
            return result

        return self.__class__.model_validate(result.model_dump_json(warnings=False))

    def encode(
        self,
        *,
        indent: int | None = None,
        include: IncEx = None,
        exclude: IncEx = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> dict[str, Any]:
        return json.loads(
            self.model_dump_json(
                indent=indent,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
            )
        )

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        extra="ignore",
    )


class TimeStampedModel(BaseModel):
    created_at: types.DateTime = field.DateTimeField()
    updated_at: types.DateTime = field.DateTimeField()



from collections.abc import Callable
from dataclasses import asdict
from typing import Annotated, Any, get_args

import pydantic_core
from fastapi import Depends, Path, Query, params
from pydantic import BaseModel, Field, Json, create_model, field_validator

from compose import container, query, types


def dict_to_json(v: dict[str, Any] | None) -> str | None:
    if v is None:
        return v

    return pydantic_core.to_json(v).decode()


TYPE_VALIDATORS = {
    Json: [
        dict_to_json,
    ]
}


def to_query[Q: BaseModel](q: type[Q], /) -> type[Q]:
    field_args = (
        "title",
        "alias",
        "default",
        "default_factory",
        "description",
    )

    validators = {}
    field_args = (*field_args, "annotation")
    field_definitions = {}
    for field_name, field_info in q.model_fields.items():
        annotation = field_info.annotation

        metadata = {}
        for item in field_info.metadata:
            metadata |= asdict(item)

        field_definitions[field_name] = (
            annotation,
            Field(
                Query(
                    **{arg: getattr(field_info, arg, None) for arg in field_args},
                    **metadata,
                )
            ),
        )
        if not (args := get_args(annotation)):
            continue

        if (arg := next((arg for arg in args if arg is not None), None)) is None:
            continue

        validators |= {
            f"{field_name}_{validator.__name__}": field_validator(field_name, mode="before")(
                validator
            )
            for validator in TYPE_VALIDATORS.get(arg, [])
        }

    return create_model(
        f"{q.__name__}Query",
        **field_definitions,
        __validators__=validators,
        __base__=q,
    )


def as_query[Q: BaseModel](q: type[Q], /) -> type[Q]:
    return Annotated[q, Depends(to_query(q))]


def create_model_dependency_resolver[T: container.BaseModel](
    model_type: type[T],
    dependencies: dict[str, tuple[type, Any]],
) -> Callable[..., Any]:
    dependencies_model = create_model(
        f"{model_type.__name__}fields",
        **{
            name: (field_type, Field(field_value))
            for name, (field_type, field_value) in dependencies.items()
        },
    )

    def wrapper(
        t: model_type,
        resolved_dependencies: Annotated[dependencies_model, Depends(dependencies_model)],
    ) -> T:
        return t.copy(update=resolved_dependencies.dict(), deep=True)

    return wrapper


# `from __future__ import annotations` 사용시 동작하지 않음
def with_depends[T: container.BaseModel](model_type: type[T], **kwargs: Any) -> type[T]:
    return Annotated[
        model_type,
        Depends(create_model_dependency_resolver(model_type, kwargs)),
    ]


class WithPath:
    @classmethod
    def object_id(
        cls, path: params.Path | None = None
    ) -> tuple[type[types.PyObjectId], params.Path]:
        return types.PyObjectId, path or Path(...)

    @classmethod
    def int(cls, path: params.Path | None = None) -> tuple[type[int], params.Path]:
        return int, path or Path(...)

    @classmethod
    def str(cls, path: params.Path | None = None) -> tuple[type[str], params.Path]:
        return str, path or Path(...)


class _OffsetPaginationParams(query.Query):
    page: int = Field(1, ge=1)
    per_page: int = Field(10, ge=1)


OffsetPaginationParams = as_query(_OffsetPaginationParams)
